package org.kantega.jexmec;

/*
 * Copyright 2009 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.simple.SimplePluginLoader;
import org.kantega.jexmec.simple.SimpleClassLoaderStrategy;

import java.util.*;
import java.lang.reflect.Proxy;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

/**
 * Default implementation of PluginManager. Tracks the discovery of ClassLoaders from the provided
 * {@link org.kantega.jexmec.ClassLoaderStrategy} implementations and uses the provided {@link PluginLoader}
 * implementations and a {@link ServiceLocator} to load plugins from the ClassLoaders
 */
public class DefaultPluginManager<P extends Plugin> implements ClassLoaderAwarePluginManager<P>, ClassLoaderListener {
    private final Collection<ClassLoaderStrategy> classLoaderStrategies = new ArrayList<ClassLoaderStrategy>();

    private final Collection<PluginLoader<P>> pluginLoaders = new ArrayList<PluginLoader<P>>();

    private final Map<ClassLoader, List<P>> dynamicPlugins = Collections.synchronizedMap(new HashMap<ClassLoader, List<P>>());

    private final ServiceLocator serviceLocator;
    private final Class<P> pluginClass;

    private enum State {STARTING, STARTED, STOPPING, STOPPED}
    private State state = State.STOPPED;

    private Map<P, ClassLoader> classLoaderMap = new HashMap<P, ClassLoader>();

    private Collection<PluginManagerListener<P>> listeners = Collections.synchronizedCollection(new ArrayList<PluginManagerListener<P>>());

    public DefaultPluginManager(Class<P> pluginClass) {
        this(pluginClass, new StaticServiceLocator());
    }

    public DefaultPluginManager(Class<P> pluginClass, ServiceLocator serviceLocator) {
        this(pluginClass, serviceLocator, null);
    }

    public DefaultPluginManager(Class<P> pluginClass, ServiceLocator serviceLocator, Class<? extends Services> servicesInterface) {
        this.pluginClass = pluginClass;
        this.serviceLocator = new PluginManagerServiceLocatorWrapper(servicesInterface == null ? serviceLocator : new ServicesClassServiceLocator(serviceLocator, servicesInterface));
    }


    public String sayHello() {return "hello";}
    public void addPluginLoader(PluginLoader<P> pluginLoader) {
        if(state != State.STOPPED) {
            throw new IllegalStateException("Illegal state " + state +" when adding plugin loader " +pluginLoader);
        }
        pluginLoaders.add(pluginLoader);
    }

    public void addClassLoaderStrategy(ClassLoaderStrategy classLoaderStrategy) {
        if(state != State.STOPPED) {
            throw new IllegalStateException("Illegal state " + state +" when adding class loader strategy " +classLoaderStrategy);
        }
        classLoaderStrategies.add(classLoaderStrategy);
    }

    public void start() {
        if(state != State.STOPPED) {
            throw new IllegalStateException("Illegal state " + state +" when starting, expected " +State.STOPPED);
        }
        state = State.STARTING;

        if(pluginLoaders.size() == 0) {
            pluginLoaders.add(createDefaultPluginLoader());
        }

        if(classLoaderStrategies.size() == 0) {
            classLoaderStrategies.add(createDefaultClassLoaderStrategy());
        }

        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.start();
        }
        for(ClassLoaderStrategy strategy : classLoaderStrategies) {
            strategy.addClassLoaderListener(this);
        }
        for(ClassLoaderStrategy strategy : classLoaderStrategies) {
            strategy.start();
        }


        state = State.STARTED;
        firePluginManagerStarted();
    }

    protected ClassLoaderStrategy createDefaultClassLoaderStrategy() {
        return new SimpleClassLoaderStrategy(getClass().getClassLoader());

    }

    protected PluginLoader<P> createDefaultPluginLoader() {
        return new SimplePluginLoader<P>();

    }

    private void firePluginManagerStarted() {
        for(PluginManagerListener listener : listeners) {
            listener.pluginManagerStarted();
        }
    }


    public void stop() {
        if(state != State.STARTED) {
            throw new IllegalStateException("Illegal state " + state +" when starting, expected " +State.STARTED);
        }
        state = State.STOPPING;
        for(ClassLoaderStrategy strategy : classLoaderStrategies) {
            strategy.stop();
        }
        for(ClassLoaderStrategy strategy : classLoaderStrategies) {
            strategy.removeClassLoaderListener(this);
        }
        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.stop();
        }


        state = State.STOPPED;
    }

    public List<P> getPlugins() {
        if(state != State.STARTED) {
            throw new IllegalStateException(DefaultPluginManager.class.getSimpleName() +" is not yet started. You need to call start() before calling getPlugins()");
        }
        List<P> plugins = new ArrayList<P>();
        for(List<P> loaderPlugins : dynamicPlugins.values()) {
            plugins.addAll(loaderPlugins);
        }
        return plugins;
    }


    public void classLoaderAdded(ClassLoader loader) {
        for(PluginManagerListener listener : listeners) {
            listener.classLoaderAdded(loader);
        }
        final List<P> loadedPlugins = loadPluginsFromClassLoader(serviceLocator, loader);
        for(P plugin : loadedPlugins) {
            classLoaderMap.put(plugin, loader);
        }

        dynamicPlugins.put(loader, loadedPlugins);
    }

    public void classLoaderRemoved(ClassLoader loader) {




        final List<P> pluginsRemoved = dynamicPlugins.get(loader);

        for(P plugin : pluginsRemoved) {
            classLoaderMap.remove(plugin);
        }
        
        dynamicPlugins.remove(loader);

        for(PluginManagerListener listener : listeners) {
            listener.classLoaderRemoved(loader);
        }
        unloadPluginsFromClassLoader(loader);

    }

    private void unloadPluginsFromClassLoader(ClassLoader loader) {
        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            pluginLoader.unloadPlugins(loader);
        }
    }

    private List<P> loadPluginsFromClassLoader(ServiceLocator serviceLocator, ClassLoader classLoader) {
        List<P> plugins = new ArrayList<P>();

        for(PluginLoader<P> pluginLoader : pluginLoaders) {
            plugins.addAll(pluginLoader.loadPlugins(pluginClass, classLoader, serviceLocator));
        }
        return plugins;
    }

    public void addPluginManagerListener(PluginManagerListener<P> pPluginManagerListener) {
        listeners.add(pPluginManagerListener);
    }

    public void removePluginManagerListener(PluginManagerListener<P> pluginManagerListener) {
        listeners.remove(pluginManagerListener);
    }

    private class ServicesClassServiceLocator implements ServiceLocator {
        private final ServiceLocator serviceLocator;
        private Services servicesProxy;
        private final Class<? extends Services> serviceClass;
        private final Set<ServiceKey> serviceClassServices;


        public ServicesClassServiceLocator(ServiceLocator serviceLocator, Class<? extends Services> serviceClass) {
            this.serviceLocator = serviceLocator;
            this.servicesProxy = createServicesProxy(serviceLocator, serviceClass);
            this.serviceClassServices = new DefaultServiceTypesProvider(serviceClass).getServiceKeys();
            this.serviceClass = serviceClass;
        }

        public <T> T lookupService(Class<T> serviceClass) {
            return lookupService(serviceClass, null);
        }

        public <T> T lookupService(Class<T> serviceClass, String name) {
            if(serviceClass ==  this.serviceClass) {
                return serviceClass.cast(servicesProxy);
            } else {
                return serviceLocator.lookupService(serviceClass, name);
            }
        }

        public Set<ServiceKey> getServiceKeys() {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            serviceKeys.add(new ServiceKey(serviceClass));
            serviceKeys.addAll(serviceClassServices);
            serviceKeys.addAll(serviceLocator.getServiceKeys());
            return serviceKeys;
        }

        private <S extends Services> S createServicesProxy(final ServiceLocator serviceLocator, final Class<S> servicesClass) {
            return servicesClass.cast(Proxy.newProxyInstance(getClass().getClassLoader(),
                    new Class[] {servicesClass},
                    new ServiceLookupInvocationHandler(serviceLocator)));
        }

    }

    private static class ServiceLookupInvocationHandler implements InvocationHandler {
            private ServiceLocator serviceLocator;

            public ServiceLookupInvocationHandler(ServiceLocator serviceLocator) {
                this.serviceLocator = serviceLocator;
            }

            public Object invoke(Object o, Method method, Object[] objects) throws Throwable {
                if(objects == null || objects.length == 0 ) {
                    return serviceLocator.lookupService(method.getReturnType());
                } else if(objects.length == 1 && objects[0] instanceof Enum) {
                    return serviceLocator.lookupService(method.getReturnType(), ((Enum)objects[0]).name());
                } else {
                    throw new IllegalArgumentException("Services interface method " + method.toString() + " should take no parameters or a single enum parameter");
                }
            }
        }

    private class PluginManagerServiceLocatorWrapper implements ServiceLocator {

        private ServiceLocator serviceLocator;

        public PluginManagerServiceLocatorWrapper(ServiceLocator serviceLocator) {

            this.serviceLocator = serviceLocator;
        }

        public <T> T lookupService(Class<T> serviceClass) {
             return lookupService(serviceClass, null);
        }

        public <T> T lookupService(Class<T> serviceClass, String name) {
            if(serviceClass == PluginManager.class) {
                return serviceClass.cast(DefaultPluginManager.this);
            } else {
                return serviceLocator.lookupService(serviceClass, name);
            }
        }

        public Set<ServiceKey> getServiceKeys() {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            serviceKeys.add(new ServiceKey(PluginManager.class));
            serviceKeys.addAll(serviceLocator.getServiceKeys());
            return serviceKeys;
        }
    }

    public ClassLoader getClassLoader(P plugin) {
        return classLoaderMap.get(plugin);
    }
}