package org.kantega.jexmec;

/*
 * Copyright 2009 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * Default implementation of {@link ServiceTypesProvider}.
 */
public class DefaultServiceTypesProvider implements ServiceTypesProvider {
    private final Set<ServiceKey> serviceClasses;

    /**
     * Use all interfaces in the provided set as service types
     * @param serviceKeys service interface classes
     */
    public DefaultServiceTypesProvider(Set<ServiceKey> serviceKeys) {
        for(ServiceKey serviceKey : serviceKeys) {
            if(!serviceKey.getServiceClass().isInterface()) {
                throw new IllegalArgumentException("Non-interface service classes are not allowed. " + serviceKey.getClass().getName() + " is not an interface");
            }
        }
        this.serviceClasses = new HashSet<ServiceKey>(serviceKeys);
    }

    /**
     * User the return types of any method in the provided {@link Services} class as service types
     * @param servicesInterface the child interface of {@link org.kantega.jexmec.Services} to
     * extract service types from.
     */
    public <S extends Services> DefaultServiceTypesProvider(Class<S> servicesInterface) {
        this(getServicesFromInterface(servicesInterface));

    }

    private static <S extends Services> Set<ServiceKey> getServicesFromInterface(Class<S> servicesInterface) {
        if (!servicesInterface.isInterface()) {
            throw new IllegalArgumentException("The class " + servicesInterface + " is not an interface");
        }

        Set<ServiceKey> serviceClasses = new HashSet<ServiceKey>();
        for (Method method : servicesInterface.getDeclaredMethods()) {
            if (method.getParameterTypes().length > 1) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " is only allowed to take a single naming enumeration as parameter. " +
                        "Please remove parameters " + Arrays.asList(method.getParameterTypes()));
            }
            Set<String> serviceNames = new HashSet<String>();

            if (method.getParameterTypes().length == 1) {
                if(!method.getParameterTypes()[0].isEnum() ) {
                    throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " a naming enumeration as a parameter. " +
                            "Please remove or fix illegal parameter " + method.getParameterTypes());
                } else {
                    Object[] constants = method.getParameterTypes()[0].getEnumConstants();
                    for(Object object : constants) {
                        if(object instanceof Enum) {
                            serviceNames.add(Enum.class.cast(object).name());
                        }

                    }
                }
            }
            if (!method.getReturnType().isInterface()) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " is not allowed to have a non-interface " +
                        "return type. " + method.getName() + " returns " + method.getReturnType() + " which is a class");
            }
            if(serviceNames.size() > 0) {
                for(String name : serviceNames) {
                    serviceClasses.add(new ServiceKey(method.getReturnType(), name));
                }
            } else {
                serviceClasses.add(new ServiceKey(method.getReturnType()));
            }
        }
        return serviceClasses;
    }

    /**
     * {@inheritDoc}
     */

    public Set<ServiceKey> getServiceKeys() {
        return serviceClasses;
    }
}
