/*
 * Copyright 2009 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.kantega.jexmec.simple;

import org.kantega.jexmec.ClassLoaderListener;
import org.kantega.jexmec.ClassLoaderStrategy;
import org.kantega.jexmec.Plugin;

import java.util.*;
import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.net.URLClassLoader;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * ClassLoaderStrategy for easy of development of plugins. On startup this strategy will set up monitoring of
 * a list of plugin directories provided through a system property given (by default {@link #DEFAULT_PLUGIN_HOMES_PROPERTY}.
 * For each of these directories the strategy will add a classloader that loads resources from the resourcesDirectory
 * (default {@link #DEFAULT_RESOURCES_DIRECTORY}) and classes from the classesDirectory (default {@link #DEFAULT_CLASSES_DIRECTORY}).
 *
 * For each directory this strategy will monitor the classes directory for changes. When a changed file is detected the
 * stratgy will remove the old classloader and add a new one. This allows for hot redeployment of single plugins.
 */
public class HotDeployingDevelopmentClassLoaderStrategy implements ClassLoaderStrategy {

    private final List<ClassLoaderListener> classLoaderListeners = Collections.synchronizedList(new ArrayList<ClassLoaderListener>());
    private static final String DEFAULT_PLUGIN_HOMES_PROPERTY = "jexmec.pluginHomes";
    private String pluginHomesProperty = DEFAULT_PLUGIN_HOMES_PROPERTY;
    private static final String DEFAULT_CLASSES_DIRECTORY = "target/classes";
    private String DEFAULT_RESOURCES_DIRECTORY = "src/main/resources";
    private String resourcesDirectory = DEFAULT_RESOURCES_DIRECTORY;
    private String classesDirectory = DEFAULT_CLASSES_DIRECTORY;
    private ClassLoader parentClassLoader = getClass().getClassLoader();
    private File workDirectory;
    private String localResourcesPrefix;
    private Thread changeDetector;
    private List<DevelopmentPluginClassLoader> currentClassLoaders = Collections.synchronizedList(new ArrayList<DevelopmentPluginClassLoader>());
    private long changeDetectorInterval = 5;
    private volatile boolean changeDetectorStopFlag = false;
    private Class<? extends Plugin> pluginClass;

    public HotDeployingDevelopmentClassLoaderStrategy(Class<? extends Plugin> pluginClass, File workDirectory) {
        this.workDirectory = workDirectory;
        this.localResourcesPrefix = "META-INF/services/" + pluginClass.getName() + "/";
        this.pluginClass = pluginClass;
    }

    public void addClassLoaderListener(ClassLoaderListener listener) {
        classLoaderListeners.add(listener);
    }

    public void removeClassLoaderListener(ClassLoaderListener listener) {
        classLoaderListeners.remove(listener);
    }

    private void notifyClassLoaderAdded(ClassLoader classLoader) {
        for (ClassLoaderListener listener : classLoaderListeners) {
            listener.classLoaderAdded(classLoader);
        }

    }

    private void notifyClassLoaderRemoved(DevelopmentPluginClassLoader classLoader) {
        for (ClassLoaderListener listener : classLoaderListeners) {
            listener.classLoaderRemoved(classLoader);
        }
    }

    public void start() {
        String homesProperty = System.getProperty(pluginHomesProperty);

        if (homesProperty != null) {
            for (String home : homesProperty.split(",;")) {
                File homeDirectory = new File(home);
                if (homeDirectory.exists()) {
                    File classes = new File(homeDirectory, classesDirectory);
                    File resources = new File(homeDirectory, resourcesDirectory);
                    if (classes.exists()) {

                        final DevelopmentPluginClassLoader loader = createClassLoader(classes, resources);
                        currentClassLoaders.add(loader);
                        notifyClassLoaderAdded(loader);
                    }

                }
            }

        }
        if(currentClassLoaders.size() > 0) {
            changeDetector = new Thread(new ChangeDetector(), "Jexmec change detector ");
            changeDetector.start();
        }

    }

    public void setLocalResourcesPrefix(String localResourcesPrefix) {
        this.localResourcesPrefix = localResourcesPrefix;
    }

    private DevelopmentPluginClassLoader createClassLoader(File classesDirectory, File resourcesDirectory) {
        return new DevelopmentPluginClassLoader(pluginClass, classesDirectory, resourcesDirectory);
    }

    public void stop() {
        changeDetectorStopFlag = true;
        changeDetector.interrupt();
    }

    private class ChangeDetector implements Runnable {


        public void run() {
            while (true) {
                try {
                    Thread.sleep(changeDetectorInterval * 1000);
                } catch (InterruptedException e) {

                }
                if(changeDetectorStopFlag) {
                    return;
                }
                List<DevelopmentPluginClassLoader> loaders = new ArrayList<DevelopmentPluginClassLoader>(currentClassLoaders);
                for (DevelopmentPluginClassLoader loader : loaders) {
                    if (loader.getLastModified() > loader.getLastCheck()) {
                        currentClassLoaders.remove(loader);
                        notifyClassLoaderRemoved(loader);
                        final DevelopmentPluginClassLoader newLoader = createClassLoader(loader.getClassesDirectory(), loader.getResourcesDirectory());
                        notifyClassLoaderAdded(newLoader);
                        currentClassLoaders.add(newLoader);
                    } else {
                        loader.setLastCheck(System.currentTimeMillis());
                    }
                }
            }
        }

    }


    private class DevelopmentPluginClassLoader extends PluginJarClassLoader {

        private URLClassLoader resourcesLoader;

        private long lastCheck = 0;

        private File classesDirectory;
        private File resourcesDirectory;

        private DevelopmentPluginClassLoader(Class<? extends Plugin> pluginClass, File classesDirectory, File resourcesDirectory) {
            super(pluginClass, classesDirectory, parentClassLoader, workDirectory);
            this.classesDirectory = classesDirectory;
            this.resourcesDirectory = resourcesDirectory;
            final List<URL> resourceUrls = new ArrayList<URL>();
            try {
                resourceUrls.add(resourcesDirectory.toURL());
            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            }

            for(int i = 1; i < getURLs().length; i++) {
                resourceUrls.add(getURLs()[i]);
            }

            resourcesLoader = new URLClassLoader(resourceUrls.toArray(new URL[resourceUrls.size()]), HotDeployingDevelopmentClassLoaderStrategy.this.parentClassLoader) {
                @Override
                public InputStream getResourceAsStream(String name) {
                    final URL resource = getResource(name);
                    try {
                        return resource == null ? null : resource.openStream();
                    } catch (IOException e) {
                        throw new RuntimeException(e);
                    }
                }

                @Override
                public URL getResource(String name) {
                    if (isLocalResource(name)) {
                        return super.findResource(name);
                    } else {
                        return super.getResource(name);
                    }
                }

                @Override
                public Enumeration<URL> getResources(String name) throws IOException {
                    if (isLocalResource(name)) {
                        return super.findResources(name);
                    } else {
                        return super.getResources(name);
                    }
                }

                protected boolean isLocalResource(String name) {
                    return name.startsWith(localResourcesPrefix);
                }

            };

            setLastCheck(getLastModified());
        }

        protected boolean isLocalResource(String name) {
            return name.startsWith(localResourcesPrefix);
        }


        @Override
        public URL getResource(String name) {
            return resourcesLoader.getResource(name);
        }

        @Override
        public Enumeration<URL> getResources(String name) throws IOException {
            return resourcesLoader.getResources(name);
        }

        @Override
        public InputStream getResourceAsStream(String name) {
            return resourcesLoader.getResourceAsStream(name);
        }

        public void setLastCheck(long lastCheck) {
            this.lastCheck = lastCheck;
        }

        public long getLastCheck() {
            return lastCheck;
        }

        public long getLastModified() {
            return getLastModified(classesDirectory);
        }

        private long getLastModified(File file) {
            if (file.isFile()) {
                return file.lastModified();
            } else {
                long max = 0;
                for (File child : file.listFiles()) {
                    max = Math.max(max, getLastModified(child));
                }
                return max;
            }
        }

        public File getClassesDirectory() {
            return classesDirectory;
        }

        public File getResourcesDirectory() {
            return resourcesDirectory;
        }
    }
}


