package org.kantega.jexmec.simple;

/*
 * Copyright 2009 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.Plugin;

import java.net.URLClassLoader;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.InputStream;
import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.util.*;
import java.util.jar.JarFile;
import java.util.jar.JarEntry;

/**
 * Class loader for loading a plugin from a jar file.
 * This class only looks for resources in the local jar file, not super classes
 */
public class PluginJarClassLoader extends URLClassLoader {
    private String localResourcePrefix;


    public PluginJarClassLoader(Class<? extends Plugin> pluginClass, File pluginJar, ClassLoader parent, File workDir) {
        super(getUrls(pluginJar, workDir), parent);
        localResourcePrefix = "META-INF/services/" + pluginClass.getName() +"/";

    }

    private static URL[] getUrls(File pluginJar, File workDir) {

        List<URL> urls = new ArrayList<URL>();


        if (pluginJar.isDirectory()) {
            try {
                urls.add(pluginJar.toURL());
                File libDir = new File(pluginJar, "META-INF/lib");
                if(libDir.exists()) {
                    for(File lib : libDir.listFiles()) {
                        urls.add(lib.toURL());
                    }
                }
            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            }
        } else {
            String pluginName = pluginJar.getName();
            pluginName = pluginName.contains(".") ? pluginName.substring(0, pluginName.lastIndexOf(".")) : pluginName;

            final File pluginWorkDir = new File(workDir, pluginName);


            try {
                urls.add(pluginJar.toURL());

                JarFile jarFile = new JarFile(pluginJar);
                List<JarEntry> entries = Collections.list(jarFile.entries());

                Set<String> libraries = new HashSet<String>();
                for (JarEntry entry : entries) {

                    if (entry.getName().startsWith("META-INF/lib") && !entry.isDirectory() && (entry.getName().endsWith(".jar"))) {
                        String depName = entry.getName().substring(entry.getName().lastIndexOf("/") + 1);

                        File expandTo = new File(pluginWorkDir, depName);


                        if (!expandTo.exists() || pluginJar.lastModified() > expandTo.lastModified()) {
                            expandTo.getParentFile().mkdirs();
                            expand(entry, jarFile, expandTo);

                        }
                        urls.add(expandTo.toURL());
                        libraries.add(depName);
                    }


                }

                // Delete stale dependencies
                if (pluginWorkDir.exists()) {
                    for (File file : pluginWorkDir.listFiles()) {
                        if (!libraries.contains(file.getName())) {
                            file.delete();
                        }
                    }
                }

                // Delete work directory if empty (no plugins)
                if (pluginWorkDir.exists() && pluginWorkDir.listFiles().length == 0) {
                    pluginWorkDir.delete();
                }

            } catch (MalformedURLException e) {
                throw new RuntimeException(e);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
        return urls.toArray(new URL[urls.size()]);
    }

    private static void expand(JarEntry entry, JarFile jarFile, File expandTo) throws IOException {
        copyStreamToFile(jarFile.getInputStream(entry), expandTo);
    }

    private static void copyStreamToFile(InputStream input, File expandTo) throws IOException {
        FileOutputStream output = null;
        try {
            output = new FileOutputStream(expandTo);
            byte[] buffer = new byte[1024 * 4];
            int read = 0;
            while (-1 != (read = input.read(buffer))) {
                output.write(buffer, 0, read);
            }
        } finally {
            if (output != null) {
                output.close();
            }
        }

    }

    @Override
    public InputStream getResourceAsStream(String name) {
        final URL resource = getResource(name);
        try {
            return resource == null ? null : resource.openStream();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public URL getResource(String name) {
        if (isLocalResource(name)) {
            return super.findResource(name);
        } else {
            return super.getResource(name);
        }
    }

    @Override
    public Enumeration<URL> getResources(String name) throws IOException {
        if (isLocalResource(name)) {
            return super.findResources(name);
        } else {
            return super.getResources(name);
        }
    }

    protected boolean isLocalResource(String name) {
        return name.startsWith(localResourcePrefix);
    };
}

