package org.kantega.jexmec.simple;

/*
 * Copyright 2009 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.*;

import java.util.*;
import java.lang.reflect.*;

/**
 * Implementation of {@link PluginLoader} based on simple constructor injection.
 * Plugin classes are located by reading the file {@code META-INF/services/&lt;pluginClassName&gt;/simple.txt} from the
 * ClassLoader. Any constructor arguments are resolved as services using the {@link org.kantega.jexmec.ServiceLocator}.
 *
 * @param <P> the Plugin implementation type this instance will load
 */
public class SimplePluginLoader<P extends Plugin> implements PluginLoader<P> {

    private final ClassLocator<Class<P>> classLocator;

    public SimplePluginLoader(ClassLocator<Class<P>> classLocator) {
        this.classLocator = classLocator;
    }

    public SimplePluginLoader() {
        classLocator = null;
    }

    public void start() {

    }

    public void stop() {

    }

    public List<P> loadPlugins(Class<P> pluginClass, ClassLoader classLoader, ServiceLocator serviceLocator) {
        ClassLocator<Class<P>> classLocator = this.classLocator != null ? this.classLocator : new DefaultClassLocator<Class<P>>(pluginClass, "simple.txt");
        List<P> plugins = new ArrayList<P>();

        final List<Class<P>> pluginClasses = classLocator.locateClasses(classLoader);


        for (Class clazz : pluginClasses) {
            if (!pluginClass.isAssignableFrom(clazz)) {
                throw new IllegalPluginClassException("Plugin class " + clazz.getName() + " is not an instance of plugin class " + pluginClass.getName(), clazz);
            }
            try {
                if (clazz.getDeclaredConstructors().length != 1) {
                    throw new MultipleConstructorsException("Simple plugin class " + clazz.getName() + " must have one and only one constructor", clazz);
                }
                Constructor constructor = clazz.getDeclaredConstructors()[0];
                Object[] params = new Object[constructor.getParameterTypes().length];
                for (int i = 0; i < constructor.getParameterTypes().length; i++) {
                    Class<?> paramClass = constructor.getParameterTypes()[i];
                    if(paramClass == NamedService.class) {
                        Type[] genericParameterTypes = constructor.getGenericParameterTypes();
                        ParameterizedType type = (ParameterizedType) genericParameterTypes[i];

                        if(type.getActualTypeArguments().length != 1) {
                            throw new IllegalArgumentException("Plugin  class " + clazz.getName() + " has an illegal constructor. " +
                                    "NamedService<?> must have exactly one actual type parameter.");

                        }

                        params[i] = new SimpleNamedService(serviceLocator, (Class) type.getActualTypeArguments()[0]);

                    } else {
                        if (!serviceLocator.getServiceKeys().contains(new ServiceKey(paramClass))) {
                            throw new UnknownServiceException("Plugin  class " + clazz.getName() + " has an illegal constructor. Parameter " + i + " of type " + paramClass.getName() + " could not be resolved to an application service", clazz, paramClass);
                        }
                        params[i] = serviceLocator.lookupService(paramClass);
                    }
                }

                P plugin = pluginClass.cast(constructor.newInstance(params));
                plugins.add(plugin);
            } catch (InstantiationException e) {
                final String msg = "Plugin class " + clazz.getName() + " is an " + (clazz.isInterface() ? "interface" : "abstract class");
                throw new IllegalPluginClassException(msg, e, clazz);
            } catch (IllegalAccessException e) {
                throw new PluginLoadingException(e);
            } catch (InvocationTargetException e) {
                throw new PluginLoadingException(e);
            }
        }
        return plugins;
    }

    public void unloadPlugins(ClassLoader classLoader) {

    }

    class SimpleNamedService implements NamedService {
        private ServiceLocator serviceLocator;
        private Class serviceClass;

        SimpleNamedService(ServiceLocator serviceLocator, Class serviceClass) {
            this.serviceLocator = serviceLocator;
            this.serviceClass = serviceClass;
        }

        public Object get(String name) {
            return serviceLocator.lookupService(serviceClass, name);
        }
    }

}
